@@ -23,7 +23,12 @@ if sys.platform == "win32":
 if sys.platform == "win32":
 	import win32api
 
+import ICCProfile as ICCP
+import colormath
+import cubeiterator as ci
 import localization as lang
+import worker_base
+from imfile import tiff_get_header
 from log import safe_print as log_safe_print
 from meta import name as appname, version
 from network import get_network_addr, get_valid_host
@@ -35,6 +40,16 @@ CALLBACK = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.POINTER(None),
 							ctypes.c_char_p, ctypes.c_ulong, ctypes.c_ulonglong,
 							ctypes.c_char_p, ctypes.c_ulonglong, ctypes.c_bool)
 
+H3D_HEADER = ("3DLT\x01\x00\x00\x00DisplayCAL\x00\x00\x00\x00\x00\x00\x00\x00"
+			  "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00 \x00"
+			  "\x00\x00\x00\x00\x00\x08\x00\x00\x00\x08\x00\x00\x00\x08\x00\x00"
+			  "\x00\x00\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00"
+			  "\x00\x00\x00\x00\x00\x00@\x00\x00\x00\x00\x00\x00\x00\x00\x00"
+			  "\x06\x00\x00\x00\x06")
+
+H3D_PARAMETERS = ("Input_Primaries %.6f %.6f %.6f %.6f %.6f %.6f %.6f %.6f\r\n"
+				  "Input_Range 16 235\r\nOutput_Range 16 235\r\n\x00")
+
 
 min_version = (0, 88, 14, 0)
 
@@ -65,8 +80,8 @@ _methodnames = ("ConnectEx", "Disable3dlut", "Enable3dlut", "EnterFullscreen",
 				"ShowProgressBar", "SetProgressBarPos",
 				"SetSelected3dlut", "SetStayOnTopButton",
 				"SetUseFullscreenButton", "ShowRGB",
-				"ShowRGBEx", "Load3dlutFile", "Disconnect",
-				"Quit", "Load3dlutFromArray256")
+				"ShowRGBEx", "Load3dlutFile", "LoadHdr3dlutFile", "Disconnect",
+				"Quit", "Load3dlutFromArray256", "LoadHdr3dlutFromArray256")
 
 _autonet_methodnames = ("AddConnectionCallback", "Listen", "Announce")
 
@@ -78,6 +93,129 @@ def safe_print(*args):
 		log_safe_print(*args)
 
 
+def icc_device_link_to_madvr(icc_device_link_filename, unity=False,
+							 colorspace=None, hdr=None, logfile=sys.stdout,
+							 convert_video_rgb_to_clut65=False):
+	"""
+	Convert ICC device link profile to madVR 256^3 3D LUT using interpolation
+	
+	madvr 3D LUT will be written to:
+	<device link filename without extension> + '.3dlut'
+	
+	"""
+	t = time()
+
+	filename, ext = os.path.splitext(icc_device_link_filename)
+
+	h3d_params = H3D_PARAMETERS
+
+	if filename.endswith(".HDR") or hdr == 2:
+		name = os.path.splitext(filename)[0]
+		h3d_params = "Input_Transfer_Function PQ\r\nOutput_Transfer_Function PQ\r\n" + h3d_params
+	elif filename.endswith(".HDR2SDR") or hdr == 1:
+		name = os.path.splitext(filename)[0]
+		h3d_params = "Input_Transfer_Function PQ\r\n" + h3d_params
+	else:
+		name = filename
+
+	if not colorspace:
+		colorspace = os.path.splitext(name)[1]
+		colorspace = colorspace[1:]
+
+	if not isinstance(colorspace, (list, tuple)):
+		key = {"BT709": "Rec. 709",
+			   "SMPTE_C": "SMPTE-C",
+			   "EBU_PAL": "PAL/SECAM",
+			   "BT2020": "Rec. 2020",
+			   "DCI_P3": "DCI P3 D65"}.get(colorspace)
+		if not key:
+			if not colorspace:
+				safe_print("ERROR - no target color space suffix in filename")
+			else:
+				safe_print("ERROR - invalid target color space:", colorspace)
+			safe_print("Possible target color spaces:",
+					   "BT709, SMPTE_C, EBU_PAL, BT2020, DCI_P3")
+			return False
+
+		rgb_space = colormath.get_rgb_space(key)
+
+		colorspace = colormath.get_rgb_space_primaries_wp_xy(rgb_space)
+
+	rx, ry, gx, gy, bx, by, wx, wy = colorspace
+
+	# Create madVR 3D LUT
+	h3d_stream = StringIO(H3D_HEADER)
+	h3dlut = H3DLUT(h3d_stream, check_lut_size=False)
+	h3dlut.parametersData = h3d_params % (rx, ry, gx, gy, bx, by, wx, wy)
+	h3dlut.write(filename + ".3dlut")
+	raw = open(filename + ".3dlut", "r+b")
+	raw.seek(h3dlut.lutFileOffset)
+	# Make sure no longer needed h3DLUT instance can be garbage collected
+	del h3dlut
+
+	# Lookup 256^3 values through device link and fill madVR cLUT
+	clutres = 256
+	clutmax = clutres - 1.0
+	if unity:
+		logfile.write("Writing unity madVR 3D LUT...\n")
+		prevperc = -1
+		for a in xrange(clutres):
+			for b in xrange(clutres):
+				for c in xrange(clutres):
+					# Optimize for speed
+					B, G, R = chr(c), chr(b), chr(a)
+					raw.write(B + B + G + G + R + R)
+			perc = round(a / clutmax * 100)
+			if perc > prevperc:
+				logfile.write("\r%i%%" % perc)
+				prevperc = perc
+	else:
+		link = ICCP.ICCProfile(icc_device_link_filename)
+		# icclu verbose=0 gives a speed increase
+		xicclu = worker_base.MP_Xicclu(link, scale=clutmax, use_icclu=True,
+									   logfile=logfile,
+									   output_format=("<H", 65535),
+									   reverse=True, output_stream=raw,
+									   convert_video_rgb_to_clut65=convert_video_rgb_to_clut65,
+									   verbose=0)
+		xicclu._in = ci.Cube3DIterator(clutres)
+		logfile.write("Looking up 256^3 input values through device link and "
+					  "writing madVR 3D LUT...\n")
+		xicclu.exit()
+		xicclu.get()
+
+	# Append a MadVR cal1 table to the 3dlut.
+	# This can be used to ensure that the Graphics Card VideoLuts
+	# are correctly setup to match what the 3dLut is expecting.
+	#
+	# Note that the calibration curves are full range, never TV encoded output values
+	#
+	# Format is (little endian):
+	#	4 byte magic number 'cal1'
+	#	4 byte version = 1
+	#	4 byte number per channel entries = 256
+	#	4 byte bytes per entry = 2
+	#	[3][256] 2 byte entry values. Tables are in RGB order
+
+	raw.write("cal1")
+	raw.write(struct.pack('<I', 1))
+	raw.write(struct.pack('<I', 256))
+	raw.write(struct.pack('<I', 2))
+	# Linear (unity) calibration
+	for i in xrange(3):
+		for j in xrange(256):
+			raw.write(struct.pack('<H', j * 257))
+	raw.close()
+
+	safe_print("")
+	safe_print("Finished up-interpolating device link in", time() - t, "seconds")
+	if filename.endswith(".HDR"):
+		safe_print("Gamut (rx ry gx gy bx by wx wy):",
+				   "%.4f %.4f %.4f %.4f %.4f %.4f %.4f %.4f" %
+				   (rx, ry, gx, gy, bx, by, wx, wy))
+	return True
+
+
 def inet_pton(ip_string):
 	"""
 	inet_pton(string) -> packed IP representation
@@ -110,10 +248,16 @@ class H3DLUT(object):
 
 	# https://sourceforge.net/projects/thr3dlut
 
-	def __init__(self, filename):
-		self.fileName = filename
-		with open(filename, "rb") as lut:
-			data = lut.read()
+	def __init__(self, stream_or_filename=None, check_lut_size=True):
+		if not stream_or_filename:
+			return
+		if isinstance(stream_or_filename, basestring):
+			self.fileName = stream_or_filename
+			with open(stream_or_filename, "rb") as lut:
+				data = lut.read()
+		else:
+			self.fileName = None
+			data = stream_or_filename.read()
 		self.signature = data[:4]
 		self.fileVersion = struct.unpack("<l", data[4:8])[0]
 		self.programName = data[8:40].rstrip("\0")
@@ -136,7 +280,7 @@ class H3DLUT(object):
 								   parametersSize]
 		self.LUTDATA = data[self.lutFileOffset:
 							self.lutFileOffset + self.lutCompressedSize]
-		if len(self.LUTDATA) != self.lutCompressedSize:
+		if check_lut_size and len(self.LUTDATA) != self.lutCompressedSize:
 			raise ValueError("3DLUT size %i does not match expected size %i" %
 							 (len(self.LUTDATA), self.lutCompressedSize))
 		if len(data) == self.lutFileOffset + self.lutCompressedSize + 1552:
@@ -170,20 +314,132 @@ class H3DLUT(object):
 	def parametersSize(self):
 		return len(self.parametersData)
 
+	def _get_stream(self, stream_or_filename=None, ext=None):
+		if not stream_or_filename:
+			stream_or_filename = self.fileName
+			if ext:
+				stream_or_filename = os.path.splitext(stream_or_filename)[0] + ext
+		if isinstance(stream_or_filename, basestring):
+			stream = open(stream_or_filename, "wb")
+		else:
+			stream = stream_or_filename
+		return stream
+
 	def write(self, stream_or_filename=None):
 		"""
 		Write 3D LUT to stream or filename.
 		
 		"""
-		if not stream_or_filename:
-			stream_or_filename = self.fileName
+		stream = self._get_stream(stream_or_filename)
+		stream.write(self.data)
 		if isinstance(stream_or_filename, basestring):
-			stream = open(stream_or_filename, "wb")
 			if not self.fileName:
 				self.fileName = stream_or_filename
+			stream.close()
+
+	def write_devicelink(self, stream_or_filename=None):
+		"""
+		Write 3D LUT to ICC device link.
+		
+		"""
+		stream = self._get_stream(stream_or_filename, ".icc")
+
+		link = ICCP.ICCProfile()
+		link.connectionColorSpace = "RGB"
+		link.profileClass = "link"
+		link.tags.desc = ICCP.TextDescriptionType()
+		link.tags.desc.ASCII = os.path.splitext(os.path.basename(stream.name))[0]
+		link.tags.cprt = ICCP.TextType("text\0\0\0\0No copyright", "cprt")
+
+		input_grid_steps = 2 ** self.inputBitDepth[0]  # Assume equal bitdepth for R, G, B
+		if input_grid_steps > 255:
+			# madVR 3D LUTs are 256^3, but ICC LUT16Type only supports up to
+			# 255^3. As madVR 3D LUTs use video levels encoding, we simply skip
+			# the first cLUT entry in each dimension and fix the offset by
+			# scaling the input/output shaper curves. That way, only level 1 of
+			# 255 will be affected (with black at 16 and white at 235),
+			# which isn't used in actual video content.
+			clut_grid_steps = 255
 		else:
-			stream = stream_or_filename
-		stream.write(self.data)
+			clut_grid_steps = input_grid_steps
+		# Filling a 255^3 list is VERY memory intensive in Python, so we 'fake'
+		# the LUT16Type cLUT and only use tag data of offsets/sizes and shaper
+		# curves while writing the raw cLUT data directly without going through
+		# decoding/re-encoding roundtrip
+		A2B0 = ICCP.LUT16Type()
+		A2B0.matrix = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
+		A2B0.input = []
+		for i in xrange(3):
+			A2B0.input.append([])
+			for j in xrange(4096):
+				A2B0.input[-1].append(min(max(j / 4095. * (256 / 255.) - (256 / 255. - 1), 0), 1) * 65535)
+		input_bytes = len(A2B0.input) * len(A2B0.input[0]) * 2
+		A2B0.clut = [[[0] * 3 for i in xrange(clut_grid_steps)]]  # Fake cLUT
+		A2B0.output = []
+		for i in xrange(3):
+			A2B0.output.append([])
+			for j in xrange(4096):
+				A2B0.output[-1].append(min(max(j / 4095. * (256 / 255.), 0), 1) * 65535)
+		output_bytes = len(A2B0.output) * len(A2B0.output[0]) * 2
+		tagData = A2B0.tagData[:52 + input_bytes]  # Exclude cLUT and output curves
+
+		# Write actual cLUT
+		# XXX Currently only 16 bit RGB data is supported
+		samples_per_pixel = 3  # RGB
+		bytes_per_sample = self.outputBitDepth / 8
+		bytes_per_pixel = samples_per_pixel * bytes_per_sample
+		io = StringIO(tagData)
+		io.seek(0, 2)  # Position cursor at end
+		i = 0
+		for R in xrange(input_grid_steps):
+			if not R:
+				i += input_grid_steps * input_grid_steps
+				continue
+			for G in xrange(input_grid_steps):
+				if not G:
+					i += input_grid_steps
+					continue
+				for B in xrange(input_grid_steps):
+					if not B:
+						i += 1
+						continue
+					index = i * samples_per_pixel * bytes_per_sample
+					BGR = self.LUTDATA[index:index + bytes_per_pixel]
+					RGB = BGR[::-1]  # BGR little-endian to RGB big-endian byte order
+					io.write(RGB)
+					i += 1
+		io.write(A2B0.tagData[-output_bytes:])  # Append output curves
+		io.seek(0)
+		link.tags.A2B0 = ICCP.ICCProfileTag(io.read(), "A2B0")
+
+		link.write(stream)
+
+		if isinstance(stream_or_filename, basestring):
+			stream.close()
+
+	def write_tiff(self, stream_or_filename=None):
+		"""
+		Write 3D LUT to TIFF file.
+		
+		"""
+		stream = self._get_stream(stream_or_filename, ".tif")
+
+		# Write image data
+		# XXX Currently only 8 or 16 bit RGB data is supported
+		samples_per_pixel = 3  # RGB
+		bytes_per_sample = self.outputBitDepth / 8
+		bytes_per_pixel = samples_per_pixel * bytes_per_sample
+		w = 2 ** self.inputBitDepth[0]  # Assume equal bitdepth for R, G, B
+		h = w * w
+		stream.write(tiff_get_header(w, h, samples_per_pixel,
+									 self.outputBitDepth))
+		entries = self.lutUncompressedSize / samples_per_pixel / bytes_per_sample
+		for i in xrange(entries):
+			index = i * samples_per_pixel * bytes_per_sample
+			BGR = self.LUTDATA[index:index + bytes_per_pixel]
+			RGB = BGR[::-1]  # BGR little-endian to RGB big-endian byte order
+			stream.write(RGB)
+
 		if isinstance(stream_or_filename, basestring):
 			stream.close()
 
@@ -201,9 +457,10 @@ class MadTPGBase(object):
 	def send(self, rgb=(0, 0, 0), bgrgb=(0, 0, 0), bits=None,
 			 use_video_levels=None, x=0, y=0, w=1, h=1):
 		cfg = self.get_pattern_config()
-		self.set_pattern_config(int(round((w + h) / 2.0 * 100)),
-								int(round(sum(bgrgb) / 3.0 * 100)), cfg[2],
-								cfg[3])
+		if cfg:
+			self.set_pattern_config(int(round((w + h) / 2.0 * 100)),
+									int(round(sum(bgrgb) / 3.0 * 100)), cfg[2],
+									cfg[3])
 		self.show_rgb(*rgb + bgrgb)
 
 
@@ -216,6 +473,8 @@ class MadTPG(MadTPGBase):
 		self._connection_callbacks = []
 
 		# We only expose stuff we might actually use.
+		# Also, as the HDR 3D LUT install API of madVR is relatively recent
+		# (September 2017), we do not require it.
 
 		# Find madHcNet32.dll
 		clsid = "{E1A8B82A-32CE-4B0D-BE0D-AA68C772E423}"
@@ -225,7 +484,12 @@ class MadTPG(MadTPGBase):
 			value, valuetype = _winreg.QueryValueEx(key, "")
 		except:
 			raise RuntimeError(lang.getstr("madvr.not_found"))
-		self.dllpath = os.path.join(os.path.split(value)[0], "madHcNet32.dll")
+		if platform.architecture()[0] == "64bit":
+			bits = 64
+		else:
+			bits = 32
+		self.dllpath = os.path.join(os.path.split(value)[0],
+									"madHcNet%i.dll" % bits)
 		if not value or not os.path.isfile(self.dllpath):
 			raise OSError(lang.getstr("not_found", self.dllpath))
 		handle = win32api.LoadLibrary(self.dllpath)
@@ -240,11 +504,19 @@ class MadTPG(MadTPGBase):
 					prefix = "AutoNet"
 				else:
 					prefix = "madVR"
-				getattr(self.mad, prefix + "_" + methodname).restype = ctypes.c_bool
+				method = getattr(self.mad, prefix + "_" + methodname, None)
+				if not method and not methodname.startswith("LoadHdr3dlut"):
+					raise AttributeError(prefix + "_" + methodname)
+				method.restype = ctypes.c_bool
 
 			# Set expected argument types
 			self.mad.madVR_ShowRGB.argtypes = [ctypes.c_double] * 3
 			self.mad.madVR_ShowRGBEx.argtypes = [ctypes.c_double] * 6
+			if hasattr(self.mad, "madVR_LoadHdr3dlutFile"):
+				self.mad.madVR_LoadHdr3dlutFile.argtypes = [ctypes.wintypes.LPWSTR,
+															ctypes.wintypes.BOOL,
+															ctypes.c_int,
+															ctypes.c_bool]
 		except AttributeError:
 			raise RuntimeError(lang.getstr("madhcnet.outdated",
 										   tuple(reversed(os.path.split(self.dllpath))) +
@@ -408,6 +680,8 @@ class MadTPG_Net(MadTPGBase):
 				sock.bind(("", port))
 				sock.listen(1)
 				thread = threading.Thread(target=self._conn_accept_handler,
+										  name="madVR.ConnectionHandler[%s]" %
+											   port,
 										  args=(sock, "", port))
 				self._threads.append(thread)
 				thread.start()
@@ -424,6 +698,8 @@ class MadTPG_Net(MadTPGBase):
 			try:
 				sock.bind(("", port))
 				thread = threading.Thread(target=self._cast_receive_handler,
+										  name="madVR.BroadcastHandler[%s:%s]" %
+											   (self.broadcast_ip, port),
 										  args=(sock, self.broadcast_ip, port))
 				self._threads.append(thread)
 				thread.start()
@@ -444,6 +720,8 @@ class MadTPG_Net(MadTPGBase):
 			try:
 				sock.bind(("", port))
 				thread = threading.Thread(target=self._cast_receive_handler,
+										  name="madVR.MulticastHandler[%s:%s]" %
+											   (self.multicast_ip, port),
 										  args=(sock, self.multicast_ip, port))
 				self._threads.append(thread)
 				thread.start()
@@ -514,6 +792,8 @@ class MadTPG_Net(MadTPGBase):
 				else:
 					self._client_sockets[addr] = conn
 					thread = threading.Thread(target=self._receive_handler,
+											  name="madVR.Receiver[%s:%s]" %
+												   addr[:2],
 											  args=(addr, conn, ))
 					self._threads.append(thread)
 					thread.start()
@@ -647,6 +927,8 @@ class MadTPG_Net(MadTPGBase):
 							else:
 								conn = self._get_client_socket(addr[0], c_port)
 								threading.Thread(target=self._connect,
+												 name="madVR.ConnectToInstance[%s:%s]" %
+													  (addr[0], c_port),
 												 args=(conn, addr[0], c_port)).start()
 					else:
 						self._casts.remove(addr)
@@ -763,6 +1045,8 @@ class MadTPG_Net(MadTPGBase):
 		for port in self.server_ports:
 			conn = self._get_client_socket(ip, port)
 			threading.Thread(target=self._connect,
+							 name="madVR.ConnectToInstance[%s:%s]" %
+								  (ip, port),
 							 args=(conn, ip, port, timeout / 1000.0)).start()
 		return self._wait_for_client((ip, port), timeout / 1000.0)
 
@@ -793,6 +1077,8 @@ class MadTPG_Net(MadTPGBase):
 				safe_print("MadTPG_Net: Connected to %s:%s" % (host, port))
 			sock.settimeout(0)
 			thread = threading.Thread(target=self._receive_handler,
+									  name="madVR.Receiver[%s:%s]" %
+										   (host, port),
 									  args=((host, port), sock, ))
 			self._threads.append(thread)
 			thread.start()
@@ -1172,17 +1458,19 @@ class MadTPG_Net_Sender(object):
 		self.command = command
 
 	def __call__(self, *args, **kwargs):
-		if self.command == "Load3dlutFile":
+		if self.command in ("Load3dlutFile", "LoadHdr3dlutFile"):
 			lut = H3DLUT(args[0])
 			lutdata = lut.LUTDATA
-			self.command = "Load3dlut"
-		elif self.command == "Load3dlutFromArray256":
+			self.command = self.command[:-4]  # Strip 'File' from command name
+		elif self.command in ("Load3dlutFromArray256", "LoadHdr3dlutFromArray256"):
 			lutdata = args[0]
-			self.command = "Load3dlut"
-		if self.command == "Load3dlut":
+			self.command = self.command[:-12]  # Strip 'File' from command name
+		if self.command in ("Load3dlut", "LoadHdr3dlut"):
 			params = struct.pack("<i", args[1])  # Save to settings?
 			params += struct.pack("<i", args[2])  # 3D LUT slot
 			params += lutdata
+			if self.command == "LoadHdr3dlut":
+				params += struct.pack("<i", args[3])  # HDR to SDR?
 		elif self.command == "SetDeviceGammaRamp":
 			params = ""
 			for j in xrange(3):
@@ -1218,7 +1506,7 @@ class MadTPG_Net_Sender(object):
 				bgb = args[5]
 			rgb = r, g, b
 			if not None in (bgr, bgg, bgb):
-				command += "Ex"
+				self.command += "Ex"
 				rgb += (bgr, bgg, bgb)
 			if None in (r, g, b):
 				raise TypeError("show_rgb() takes at least 4 arguments (%i given)" %
